#include "indigo.h"
#include <stdio.h>
#include <stdlib.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

static void reportError(int error_code) {
  fprintf(stderr,"Occam API Error: %i\n",error_code);
  abort();
}

bool getParamRange(OccamDevice* device, OccamDeviceParam param_type, int& min_value, int& max_value) {
  int r;
  OccamDeviceParamList* param_list;
  if ((r = occamEnumerateDeviceParamList(device, &param_list)) != OCCAM_API_SUCCESS)
    return false;
  for (int j=0;j<param_list->param_count;++j) {
    if (param_list->params[j].id == param_type) {
      min_value = int(param_list->params[j].min_value);
      max_value = int(param_list->params[j].max_value);
      occamFreeDeviceParamList(param_list);
      return true;
    }
  }
  occamFreeDeviceParamList(param_list);
  return false;
}

int main(int argc, const char** argv) {
  int r;
  int i;
  int dev_index = argc>=2?atoi(argv[1]):0;
  OccamDeviceList* device_list;
  OccamDevice* device;
  OccamImage* image;

  if ((r = occamInitialize()) != OCCAM_API_SUCCESS)
    reportError(r);

  if ((r = occamEnumerateDeviceList(2000, &device_list)) != OCCAM_API_SUCCESS)
    reportError(r);
  printf("%i devices found\n", device_list->entry_count);
  for (i=0;i<device_list->entry_count;++i) {
    printf("device[%i]: cid = %s\n",
	   i,device_list->entries[i].cid);
  }
  if (dev_index<0 || dev_index >= device_list->entry_count) {
    fprintf(stderr,"device index %i out of range\n",dev_index);
    return 1;
  }

  if ((r = occamOpenDevice(device_list->entries[dev_index].cid, &device)) != OCCAM_API_SUCCESS)
    reportError(r);

  int min_exposure;
  int max_exposure;
  int min_gain;
  int max_gain;
  if (!getParamRange(device, OCCAM_EXPOSURE, min_exposure, max_exposure)) {
    fprintf(stderr,"failed to discover exposure range\n");
    return 1;
  }
  if (!getParamRange(device, OCCAM_GAIN, min_gain, max_gain)) {
    fprintf(stderr,"failed to discover exposure range\n");
    return 1;
  }
  printf("exposure range: [%i, %i]\n", min_exposure, max_exposure);
  printf("gain range: [%i, %i]\n", min_gain, max_gain);
  int exposure = (max_exposure + min_exposure) / 2;
  int gain = (max_gain + min_gain) / 2;
  bool downsample = true;

  printf("COMMANDS:\n");
  printf("  ESC/Q/q: quit this application\n");
  printf("  s: disable stitching\n");
  printf("  d: enable stitching\n");
  printf("  e: disable auto-exposure\n");
  printf("  r: enable auto-exposure\n");
  printf("  g: disable auto-gain\n");
  printf("  h: enable auto-gain\n");
  printf("  [: decrease exposure\n");
  printf("  ]: increase exposure\n");
  printf("  ;: decrease gain\n");
  printf("  ': increase gain\n");
  printf("  w: toggle downsampling to 1024-width\n");

  for (i=0;;++i) {
    if ((r = occamDeviceReadImage(device, &image, 1)) != OCCAM_API_SUCCESS)
      reportError(r);

    //    printf("read image %i: time_ns = %llu, index = %i, format = %i, width = %i, height = %i\n",
    //    	   i, image->time_ns, image->index, image->format, image->width, image->height);

    cv::Mat img;
    if (image->format == OCCAM_GRAY8)
      img = cv::Mat_<uchar>(image->height,image->width,(uchar*)image->data[0],image->step[0]);
    else if (image->format == OCCAM_RGB24)
      img = cv::Mat_<cv::Vec3b>(image->height,image->width,(cv::Vec3b*)image->data[0],image->step[0]);

    if (downsample) {
      int small_width = 1024;
      int small_height = image->height * small_width / image->width;
      cv::Mat small_img;
      cv::resize(img, small_img, cv::Size(small_width,small_height));
      img = small_img;
    }

    imshow("omni60", img);

    int key = cv::waitKey(1);
    if (key == 27 || key == 'Q' || key == 'q') // ESC/Q/q keys to quit
      break;

    else if (key == 's') { // disable stitching
      occamSetDeviceValuei(device, OCCAM_STITCHING_ENABLED, 0);
      printf("stitching disabled\n");
    }
    else if (key == 'd') { // enable stitching
      occamSetDeviceValuei(device, OCCAM_STITCHING_ENABLED, 1);
      printf("stitching enabled\n");
    }

    else if (key == 'e') { // disable auto-exposure
      occamSetDeviceValuei(device, OCCAM_AUTO_EXPOSURE, 0);
      printf("auto-exposure disabled\n");
    }
    else if (key == 'r') { // enable auto-exposure
      occamSetDeviceValuei(device, OCCAM_AUTO_EXPOSURE, 1);
      printf("auto-exposure enabled\n");
    }

    else if (key == 'g') { // disable auto-gain
      occamSetDeviceValuei(device, OCCAM_AUTO_GAIN, 0);
      printf("auto-gain disabled\n");
    }
    else if (key == 'h') { // enable auto-gain
      occamSetDeviceValuei(device, OCCAM_AUTO_GAIN, 1);
      printf("auto-gain enabled\n");
    }

    else if (key == ']') { // increase exposure
      exposure = std::min(max_exposure,exposure+1);
      printf("manual exposure: %i\n", exposure);
      occamSetDeviceValuei(device, OCCAM_EXPOSURE, exposure);
    } 
    else if (key == '[') { // decrease exposure
      exposure = std::max(min_exposure,exposure-1);
      printf("manual exposure: %i\n", exposure);
      occamSetDeviceValuei(device, OCCAM_EXPOSURE, exposure);
    }

    else if (key == '\'') { // increase gain
      gain = std::min(max_gain,gain+1);
      printf("manual gain: %i\n", gain);
      occamSetDeviceValuei(device, OCCAM_GAIN, gain);
    }
    else if (key == ';') { // decrease gain
      gain = std::max(min_gain,gain-1);
      printf("manual gain: %i\n", gain);
      occamSetDeviceValuei(device, OCCAM_GAIN, gain);
    }

    else if (key == 'w') { // toggle downsampling
      downsample = !downsample;
      printf("downsampling: %i\n", downsample);
    }

    occamFreeImage(image);
  }

  occamCloseDevice(device);
  occamFreeDeviceList(device_list);
  occamShutdown();

  return 0;
}
